/* $Workfile: AlvisDmp.c $       $Revision: 16 $      $Modtime: 4/15/09 11:36a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc OtxDriverTest

  @module

    AlvisDmp.c

    Demo application for Alvis-ASM board populated one one of the following boards:
    
      Thor-8-PCI-Plus (2.0 or 1.0)
      Thor-4-PCI-Plus
      Thor-2-PCI-Plus
      Gimle-16-PCI-Plus

    The program communicates with one of the DMPs on the Alvis-ASM. The DMP number 
    is selected via a command line switch.

    The demo configures all Li (E1/T1) devices on and sends all of the E1/T1
    timeslots to the DMP.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDmp.h"
#include "OtxEth.h"

#define DEMO_DEFAULT_LOG_FILE "AlvisDmp.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_DEFAULT_DMP_NO 0

static void DemoLogf(char *formatStr, ...);
static void DemoExitQuestion();

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  OTX_INT32              nBoardType;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  int                    boardNo;
  int                    nDmp;         // 0-3
  int                    coLi; // 2, 4, 8, or 16
  int                    nLinkTimeout;  // in seconds
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_UINT16             nLineLength;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  int                    nIpAddrLsb;
  int                    nMacAddrLsb;
} DemoEnvT;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
DemoEnvT    g_demoEnv;
FILE *g_pDemoLogFile = NULL;
OTX_BOOL g_bDmpLinkInitialized = OTX_FALSE;

OTX_HANDLE g_hBoard;
OTX_HANDLE g_hAsm;
OTX_HANDLE g_hEthSw;
OTX_HANDLE g_hDmp[8];
OTX_HANDLE g_hTss;  
OTX_HANDLE g_hQuadE1T1[4];
OTX_HANDLE g_hLi[16];

struct {
  OTX_HANDLE hHdlcSender[16];
  OTX_HANDLE hHdlcReceiver[16];
} g_hLogicDevs;



/*******************************************************************************
*******************************************************************************/
void DemoCheckRc(IN OTX_RESULT result)
{
  if ( OTX_FAILURE(result) ) {
    char msgStr[256];
    OTX_RESULT otxRc;
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}

/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  DemoLogf("Do you want to exit (Y/N) ");
  c = OsaConsoleGetCh();
  
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}



/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DMP:
    // Event from the DMP device
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DMP Event: %s \n", msgStr);
    } else {
      DemoLogf("DMP Invalid Event: %ld \n", pEventData->m_nCode);
    }
    if (pEventData->m_nCode == OTX_DMP_EC_LINK_INITIALIZED) {
      g_bDmpLinkInitialized = OTX_TRUE;
    }
    break;
    
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;
    
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_RECEIVER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from HLDC Receiver
      // @todo DSP and LI HdlcReceivers should return the same event code since they are of the same logical device type
      if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
        OTX_UINT8  buf[1024];
        OTX_UINT32 nFrameLen;
        OTX_UINT32 nFrameSeq;
        OtxHdlcFrameEndCodeE nFrameEndCode;
        OTX_DATETIME nTimeStamp;
        OTX_UINT32 i;

        DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);

        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
          DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          for (i = 0; i < nFrameLen && i < 1024; i++) {
            DemoLogf("%.2X ", buf[i]);
          }
          DemoLogf("\n");
        } else {
          DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
          DemoCheckRc(otxRc);
        }
      } else {
        // Unexpected event from this device
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  case OTX_DEVICE_ETHERNET_SWITCH:
    {
      char *szStatus;
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      switch (pEventData->m_nCode) {
      case OTX_ETH_EC_LEARNED_MAC_ADDRESS:
        // This event carry additional information so we treat it separately
        //DemoLogf("EthSw[] Event: %s (0x%.8X 0x%.8X)\n", msgStr, pEventData->m_nRequestId, pEventData->m_nParam);
        DemoLogf("EthSw[] Event: %s\n", msgStr);
        switch (pEventData->m_nParam & 0x00007000) {
        case 0x00001000:
          szStatus = "Dynamic MAC Address";
          break;
        case 0x00003000:
          szStatus = "Static MAC Address";
          break;
        case 0x00005000:
          szStatus = "Static MAC Address with source and destination filter";
          break;
        case 0x00006000:
          szStatus = "Static MAC Address with source filter";
          break;
        case 0x00007000:
          szStatus = "Static MAC Address with destination filter";
          break;
        default:
          szStatus = "Invalid or Reserved";
          break;
        }
        DemoLogf("  Status: %s\n", szStatus);
        DemoLogf("  VLANID: %d\n", pEventData->m_nParam & 0xFFF);
        DemoLogf("  MAC   : %.2X-%.2X-%.2X-%.2X-%.2X-%.2X\n", 
          (OTX_UINT8)((pEventData->m_nRequestId >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 16) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 8) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 0) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 16) & 0xFF)
          );
        break;
      case OTX_ETH_EC_DELETED_MAC_ADDRESS:
        // This event carry additional information so we treat it separately
        DemoLogf("EthSw[] Event: %s\n", msgStr);
        DemoLogf("  VLANID: %d\n", pEventData->m_nParam & 0xFFF);
        DemoLogf("  MAC   : %.2X-%.2X-%.2X-%.2X-%.2X-%.2X\n", 
          (OTX_UINT8)((pEventData->m_nRequestId >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 16) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 8) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 0) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 16) & 0xFF)
          );
        break;
      case OTX_ETH_EC_QOS_STATUS:
        DemoLogf("EthSw[Port#%d] Event: %s, QoS Status =  0x%.4X\n", pEventData->m_nParam, msgStr, (OTX_UINT16)pEventData->m_nRequestId);
        break;
      default:
        // Other events have the Ethernet Switch port number in pEventData->m_nParam
        DemoLogf("EthSw[Port#%d] Event: %s \n", pEventData->m_nParam, msgStr);
        break;
      }
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
#define DEMO_EVENT_DATA_BUF_SIZE 10
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}


/*******************************************************************************

  @func

  Print Statistical Ethernet counters

  @comm

*******************************************************************************/
void DemoPrintStatCounters(OtxEthStatCountersS *pStatCounters)
{
  //DemoLogf("   nBytesReceived             : %-8d\n",  pStatCounters->nBytesReceived            ); 
  DemoLogf("   nFramesReceived            : %-8d\n",  pStatCounters->nFramesReceived           ); 
  //DemoLogf("   nTotalBytesReceived        : %-8d\n",  pStatCounters->nTotalBytesReceived       ); 
  DemoLogf("   nTotalFramesReceived       : %-8d\n",  pStatCounters->nTotalFramesReceived      ); 
  DemoLogf("   nFlowCtrlFramesReceived    : %-8d\n",  pStatCounters->nFlowCtrlFramesReceived   ); 
  DemoLogf("   nMulticastFramesReceived   : %-8d\n",  pStatCounters->nMulticastFramesReceived  ); 
  DemoLogf("   nBroadcastFramesReceived   : %-8d\n",  pStatCounters->nBroadcastFramesReceived  ); 
  //DemoLogf("   nBytesSent                 : %-8d\n",  pStatCounters->nBytesSent                ); 
  DemoLogf("   nUnicastFramesSent         : %-8d\n",  pStatCounters->nUnicastFramesSent        ); 
  DemoLogf("   nNonUnicastFrmsSent        : %-8d\n",  pStatCounters->nNonUnicastFrmsSent       ); 
  DemoLogf("   nFlowCtrlFramesSent        : %-8d\n",  pStatCounters->nFlowCtrlFramesSent       ); 
  DemoLogf("   nFramesWithLen64           : %-8d\n",  pStatCounters->nFramesWithLen64          ); 
  DemoLogf("   nFramesLen65To127          : %-8d\n",  pStatCounters->nFramesLen65To127         ); 
  DemoLogf("   nFramesLen128To255         : %-8d\n",  pStatCounters->nFramesLen128To255        ); 
  DemoLogf("   nFramesLen256To511         : %-8d\n",  pStatCounters->nFramesLen256To511        ); 
  DemoLogf("   nFramesLen512To1023        : %-8d\n",  pStatCounters->nFramesLen512To1023       ); 
  DemoLogf("   nFramesLen1024ToMax        : %-8d\n",  pStatCounters->nFramesLen1024ToMax       ); 
  DemoLogf("   nOversizeFrames            : %-8d\n",  pStatCounters->nOversizeFrames           ); 
  DemoLogf("   nJabberFrames              : %-8d\n",  pStatCounters->nJabberFrames             ); 
  DemoLogf("   nFragments                 : %-8d\n",  pStatCounters->nFragments                ); 
  DemoLogf("   nShortEvents               : %-8d\n",  pStatCounters->nShortEvents              ); 
  DemoLogf("   nFrameTransFailures        : %-8d\n",  pStatCounters->nFrameTransFailures       ); 
  DemoLogf("   nAlignmentErrors           : %-8d\n",  pStatCounters->nAlignmentErrors          ); 
  DemoLogf("   nUndersizeFrames           : %-8d\n",  pStatCounters->nUndersizeFrames          ); 
  DemoLogf("   nCrcErrors                 : %-8d\n",  pStatCounters->nCrcErrors                ); 
  DemoLogf("   nCollisions                : %-8d\n",  pStatCounters->nCollisions               ); 
  DemoLogf("   nFramesDroppedCongestion   : %-8d\n",  pStatCounters->nFramesDroppedCongestion  ); 
  DemoLogf("   nFramesFilteredByEngine    : %-8d\n",  pStatCounters->nFramesFilteredByEngine   ); 
  DemoLogf("   nLateCollision             : %-8d\n",  pStatCounters->nLateCollision            ); 
}

/*******************************************************************************

  @func

  Print Ethernet port status

  @comm

*******************************************************************************/
void DemoPrintPortStatus(OtxEthPortStatusS *pPortStatus)
{
  DemoLogf("   Link Status : %s\n",  pPortStatus->bLinkUp ? "Up" : "Down"); 
  DemoLogf("   Link Speed  : %d\n",  pPortStatus->nLinkSpeed); 
  DemoLogf("   Flow Control: %s\n",  pPortStatus->bFlowControl ? "Yes" : "No"); 
  DemoLogf("   Full Duplex : %s\n",  pPortStatus->bFullDuplex ? "Yes" : "No"); 
}

/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd   Synopsis\n");
    DemoLogf("===   ========\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");
    DemoLogf(" b    Boot DMP[%d]\n", g_demoEnv.nDmp);
    DemoLogf(" e    Run standalone process on DMP[%d]\n", g_demoEnv.nDmp);
    DemoLogf(" E    Kill standalone process on DMP[%d]\n", g_demoEnv.nDmp);
    DemoLogf(" x    Upload a file to DMP[%d]\n", g_demoEnv.nDmp);
    DemoLogf(" X    Download a file from DMP[%d]\n", g_demoEnv.nDmp);
    DemoLogf(" R    Read 32-bit register from DMP device\n");
    DemoLogf(" W    Write 32-bit register to DMP device\n");
    DemoLogf(" d    Run OTX_DMP_SPM_OTXSPM2 and connect logical devices on DMP device\n");
    DemoLogf(" D    Run CodecEngineServer on DMP device\n");
    DemoLogf(" #    Restart CodecEngineServer on DMP device\n");
    DemoLogf(" N    Set DMP network parameters\n");
    DemoLogf(" s    Ethernet port statistics (for DMP port [%d])\n", g_demoEnv.nDmp);
    DemoLogf(" S    Ethernet port statistics (for GBE port [%d])\n", 9);
    DemoLogf(" p    Ethernet port status (for DMP port [%d])\n", g_demoEnv.nDmp);
    DemoLogf(" P    Ethernet port status (for GBE port [%d])\n", 9);
    DemoLogf(" c    Send a Control Command (Memory Read) to the Ethernet switch\n", 9);
    DemoLogf(" =    Toggle OTX_ATTR_ETH_REPORT_MAC_AND_VLAN attribute\n");
    DemoLogf(" +    Toggle OTX_ATTR_ETH_REPORT_WRAPPED_COUNTERS attribute\n");
    DemoLogf("\n");
#if 1
    DemoLogf("Commands for debugging purposes only:\n");
    DemoLogf(" ^    Run extended upload/download test\n");
    DemoLogf(" l    Turn Alvis-ASM LED off\n");
    DemoLogf(" L    Turn Alvis-ASM LED on\n");
    DemoLogf(" r    Read indirect register from Ethernet Sw device\n");
    DemoLogf(" w    Write indirect register to Ethernet Sw device\n");
    DemoLogf(" i    Read direct register from Ethernet Sw device\n");
    DemoLogf(" o    Write direct register to Ethernet Sw device\n");
    DemoLogf(" J    Read register from board device\n");
    DemoLogf(" K    Write register to board device\n");
    DemoLogf(" C    Send Constant Byte on TSS Highway\n");
    DemoLogf(" *    Reset and re-configure Ethernet Sw device\n");
    DemoLogf(" t    Transmit UDP Packet to DMP0 (destport:11000, sourceport:11000)\n");
    DemoLogf(" !    Ethernet port statistics (for CPU port [%d])\n", 8);
    DemoLogf(" |    Test (blink) PHY LEDs\n");
#endif
    break;

  case '=':
    {
      OTX_BOOL bAttrOn;

      result = OtxDrvGetAttributeValueBool(g_hEthSw, OTX_ATTR_ETH_REPORT_MAC_AND_VLAN, &bAttrOn); DemoCheckRc(result);
      bAttrOn = !bAttrOn;

      DemoLogf("Setting OTX_ATTR_ETH_REPORT_MAC_AND_VLAN attribute to %s\n", bAttrOn ? "TRUE" : "FALSE");

      result = OtxDrvSetAttributeValueBool(g_hEthSw, OTX_ATTR_ETH_REPORT_MAC_AND_VLAN, bAttrOn); DemoCheckRc(result);

      DemoLogf("Done\n");
    }
    break;

  case '+':
    {
      OTX_BOOL bAttrOn;

      result = OtxDrvGetAttributeValueBool(g_hEthSw, OTX_ATTR_ETH_REPORT_WRAPPED_COUNTERS, &bAttrOn); DemoCheckRc(result);
      bAttrOn = !bAttrOn;

      DemoLogf("Setting OTX_ATTR_ETH_REPORT_WRAPPED_COUNTERS attribute to %s\n", bAttrOn ? "TRUE" : "FALSE");

      result = OtxDrvSetAttributeValueBool(g_hEthSw, OTX_ATTR_ETH_REPORT_WRAPPED_COUNTERS, bAttrOn); DemoCheckRc(result);

      DemoLogf("Done\n");
    }
    break;

  case 'c':
    {
      OTX_INT32 i;
      OTX_UINT8 sqCmd[OTX_ETH_CTRL_CMD_FRAME_SIZE];
      OTX_UINT8 sqResponse[OTX_ETH_CTRL_CMD_FRAME_SIZE];
      OTX_UINT32 nMemAddr;
      OTX_INT32 iBlock = 0;
      IN OTX_INT32 nPort = 9;
      
      DemoLogf("Send a control command (memory read) to the Ethernet Switch\n");
      
      nMemAddr = nPort * 0x08 + iBlock*4;
      
      memset(sqCmd, 0, sizeof(sqCmd));
      memset(sqResponse, 0, sizeof(sqResponse));
      
      sqCmd[0] = 0x18;  // CMD0 (1x=Read, and x8=Statistics Counter Memory)
      sqCmd[1] = 0x00;  // CMD1 (0x00=Statistics Counter Memory)
      
      // 24-bit address
      sqCmd[2] = (OTX_UINT8)(nMemAddr >> 0 & 0xFF);  
      sqCmd[3] = (OTX_UINT8)(nMemAddr >> 8 & 0xFF);  
      sqCmd[4] = (OTX_UINT8)(nMemAddr >> 16 & 0xFF);  
      
      // Size
      sqCmd[5] = 0x04; // 0x04=Read 32 bytes
      
      // Frame
      sqCmd[6] = 0xAB; // Arbitrary frame number
      
      // Dummy
      sqCmd[7] = 0x00; // Dummy byte
      
      result = OtxEthControlCommand(g_hEthSw, sqCmd, sqResponse); DemoCheckRc(result);
      
      DemoLogf("Response:\n");
      for (i = 0; i < OTX_ETH_CTRL_CMD_FRAME_SIZE; i += 8) {
        DemoLogf("%d: %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X\n",   
          i/8,
          sqResponse[i+7], 
          sqResponse[i+6], 
          sqResponse[i+5], 
          sqResponse[i+4], 
          sqResponse[i+3], 
          sqResponse[i+2], 
          sqResponse[i+1], 
          sqResponse[i+0]
          );
      }
        
    }
    break;

  case 's':
    {
      OTX_INT32 nPort;
      OtxEthStatCountersS sStatCounters;
      
      nPort = g_demoEnv.nDmp;
      result = OtxEthGetStatCounters(g_hEthSw, nPort, &sStatCounters); DemoCheckRc(result);
      
      DemoLogf("Statistical Counters for Ethernet Switch port %d (DMP)\n", nPort);
      DemoPrintStatCounters(&sStatCounters);
    }
    break;


  case 'S':
    {
      OTX_INT32 nPort;
      OtxEthStatCountersS sStatCounters;
      
      nPort = 9;
      result = OtxEthGetStatCounters(g_hEthSw, nPort, &sStatCounters); DemoCheckRc(result);
      
      DemoLogf("Statistical Counters for Ethernet Switch port %d (GBE)\n", nPort);
      DemoPrintStatCounters(&sStatCounters);
    }
    break;

  case 'p':
    {
      OTX_INT32 nPort;
      OtxEthPortStatusS sPortStatus;
      
      nPort = g_demoEnv.nDmp;
      result = OtxEthGetPortStatus(g_hEthSw, nPort, &sPortStatus); DemoCheckRc(result);
      
      DemoLogf("Port Status for Ethernet Switch port %d (DMP)\n", nPort);
      DemoPrintPortStatus(&sPortStatus);
    }
    break;


  case 'P':
    {
      OTX_INT32 nPort;
      OtxEthPortStatusS sPortStatus;
      
      nPort = 9;
      result = OtxEthGetPortStatus(g_hEthSw, nPort, &sPortStatus); DemoCheckRc(result);
      
      DemoLogf("Port Status for Ethernet Switch port %d (GBE)\n", nPort);
      DemoPrintPortStatus(&sPortStatus);
    }
    break;


  case 'b':
    {
      OTX_INT32 c;
      OTX_HANDLE hDmp = g_hDmp[g_demoEnv.nDmp];

      DemoLogf("Boot DMP[%d]\n", g_demoEnv.nDmp);

      DemoLogf("Select Boot Source:\n");

      DemoLogf("0 = UART (mainly used to reload u-boot)\n");
      DemoLogf("1 = FLASH (normal boot source)\n");
      DemoLogf("2 = DSP Selfboot (use when debugging with Code Composer)\n");
      DemoLogf("3 = I changed my mind. I don't want to boot the DMP.\n");
      
      c = OsaConsoleGetCh();

      DemoLogf("%c pressed\n", c);

      switch (c - '0') {
      case 0:
        DemoLogf("Booting from UART\n");
        result = OtxDmpBoot(hDmp, OTX_DMP_BOOT_UART); DemoCheckRc(result);
        break;
      case 1:
        DemoLogf("Booting from FLASH\n");
        result = OtxDmpBoot(hDmp, OTX_DMP_BOOT_FLASH); DemoCheckRc(result);
        break;
      case 2:
        DemoLogf("DSP Self-boot\n");
        result = OtxDmpBoot(hDmp, OTX_DMP_BOOT_DSP); DemoCheckRc(result);
        break;
      default:
        DemoLogf("Boot operation aborted\n");
        break;
      }
    
      DemoLogf("Done\n");
    }
    break;


  case 'd':
    {
      OTX_INT32 nLi;
      OTX_INT32 nDmp;
      OTX_INT32 i = 0;
      OTX_RESULT result;
      OTX_UINT8 sqLiTimeSlotMask[32];
      OTX_UINT8 sqDmpTimeSlotMask[128];
      OTX_INT32 nTsLi;
      OTX_INT32 nTsDmp;
      OTX_UINT32 coTsPerDmpHwy;
      OTX_UINT32 nDmpHighwaySpeed;
      OTX_HANDLE hPipeHdlcRx[16]; // One per Li
      OTX_HANDLE hPipeHdlcTx[16]; // One per Li

      i = 0;
      nLi = 0;
      nDmp = g_demoEnv.nDmp;
      
      
      DemoLogf("Run OTX_DMP_SPM_OTXSPM2\n");
      result = OtxDmpRunStandardProgramModule(g_hDmp[nDmp], OTX_DMP_SPM_OTXSPM2); DemoCheckRc(result);
      
      // Create Hdlc Senders and Receivers logical devices hosted in the Dmp devices
      DemoLogf("Create HDLC Sender\n");
      result = OtxDrvCreateLogicalDevice(g_hDmp[nDmp], OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(result);
      DemoLogf("Create HDLC Receiver\n");
      result = OtxDrvCreateLogicalDevice(g_hDmp[nDmp], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(result);
      
      
      // Create an incoming and outgoing Pipe representing the external T1/E1
      result = OtxDrvCreatePipe(64, &hPipeHdlcRx[i]); DemoCheckRc(result);
      result = OtxDrvCreatePipe(64, &hPipeHdlcTx[i]); DemoCheckRc(result);
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hDmp[nDmp], OTX_ATTR_DMP_HIGHWAY_SPEED, &nDmpHighwaySpeed); 
      DemoLogf("DMP#%d highway speed is %d kbps\n", nDmp, nDmpHighwaySpeed);
      coTsPerDmpHwy = nDmpHighwaySpeed / 64;
      
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        nTsLi = 24; 
      } else {
        // E1 mode
        nTsLi = 16; 
      }
      nTsDmp = nTsLi;
      
      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      memset(sqDmpTimeSlotMask, 0, coTsPerDmpHwy * sizeof(OTX_UINT8));
      sqDmpTimeSlotMask[nTsDmp] = 0xFF;
      
      DemoLogf("Connecting HDLC device[%d]: nLi=%d, nTsLi=%.2d, nDmp=%d, nTsDmp=%.3d\n", i, nLi, nTsLi, nDmp, nTsDmp);
      
      DemoLogf("Connecting       hPipeHdlcRx[i], OTX_PIPE_INPUT,  g_hLi[nLi]\n");
      result = OtxDrvConnectPipe(hPipeHdlcRx[i], OTX_PIPE_INPUT,  g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);

      DemoLogf("Connecting       hPipeHdlcRx[i], OTX_PIPE_OUTPUT, g_hDmp[nDmp]\n");
      result = OtxDrvConnectPipe(hPipeHdlcRx[i], OTX_PIPE_OUTPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, coTsPerDmpHwy); DemoCheckRc(result);
      
       // Connect HDLC Sender and Receiver devices to the pipes
      DemoLogf("Connecting Logical Devices within the DMP (RX)\n");
      result = OtxDrvConnectLogicalDevice(hPipeHdlcRx[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(result);

      DemoLogf("Connecting       hPipeHdlcTx[i], OTX_PIPE_OUTPUT, g_hLi[nLi]\n");
      result = OtxDrvConnectPipe(hPipeHdlcTx[i], OTX_PIPE_OUTPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);

      DemoLogf("Connecting       hPipeHdlcTx[i], OTX_PIPE_INPUT, g_hDmp[nDmp]\n");
      result = OtxDrvConnectPipe(hPipeHdlcTx[i], OTX_PIPE_INPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, coTsPerDmpHwy); DemoCheckRc(result);
      
      // Connect HDLC Sender and Receiver devices to the pipes
      DemoLogf("Connecting Logical Devices within the DMP (TX)\n");
      result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, hPipeHdlcTx[i], -1); DemoCheckRc(result);

      // Initialize the Hdlc Sender and Receiver
      DemoLogf("Initialize the Hdlc Sender and Receiver\n");
      result = OtxHdlcSenderInit(
        g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
        2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                    // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
        OTX_HDLC_TR_NA);             // HDLC Transmission Rate for V.110 and X.30 modes
      DemoCheckRc(result);
      
      result = OtxHdlcReceiverInit(
        g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
        2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
        OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
        OTX_TRUE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      DemoCheckRc(result);
      
      result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      
      DemoLogf("Done\n");
    }
    break;

#if 1 // Just for testing
  case '@':
    {
      OTX_INT32 nLi;
      OTX_INT32 nDmp;
      OTX_INT32 i = 0;
      OTX_RESULT result;
      OTX_UINT8 sqLiTimeSlotMask[32];
      OTX_UINT8 sqDmpTimeSlotMask[128];
      OTX_INT32 nTsLi;
      OTX_INT32 nTsDmp;
      OTX_UINT32 coTsPerDmpHwy;
      OTX_UINT32 nDmpHighwaySpeed;
      OTX_HANDLE hPipeHdlcRx[8]; // One per Li
      OTX_HANDLE hPipeHdlcTx[8]; // One per Li

      i = 0;
      nLi = 0;
      nDmp = g_demoEnv.nDmp;
      
      
      // Create an incoming and outgoing Pipe representing the external T1/E1
      result = OtxDrvCreatePipe(64, &hPipeHdlcRx[i]); DemoCheckRc(result);
      result = OtxDrvCreatePipe(64, &hPipeHdlcTx[i]); DemoCheckRc(result);
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hDmp[nDmp], OTX_ATTR_DMP_HIGHWAY_SPEED, &nDmpHighwaySpeed); 
      DemoLogf("DMP#%d highway speed is %d kbps\n", nDmp, nDmpHighwaySpeed);
      coTsPerDmpHwy = nDmpHighwaySpeed / 64;
      
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        nTsLi = 24; 
      } else {
        // E1 mode
        nTsLi = 16; 
      }
      nTsDmp = nTsLi;
      
      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      memset(sqDmpTimeSlotMask, 0, coTsPerDmpHwy * sizeof(OTX_UINT8));
      sqDmpTimeSlotMask[nTsDmp] = 0xFF;
      
      DemoLogf("Connecting HDLC device[%d]: nLi=%d, nTsLi=%.2d, nDmp=%d, nTsDmp=%.3d\n", i, nLi, nTsLi, nDmp, nTsDmp);
      
      DemoLogf("Connecting       hPipeHdlcRx[i], OTX_PIPE_INPUT,  g_hLi[nLi]\n");
      result = OtxDrvConnectPipe(hPipeHdlcRx[i], OTX_PIPE_INPUT,  g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);

      DemoLogf("Connecting       hPipeHdlcRx[i], OTX_PIPE_OUTPUT, g_hDmp[nDmp]\n");
      result = OtxDrvConnectPipe(hPipeHdlcRx[i], OTX_PIPE_OUTPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, coTsPerDmpHwy); DemoCheckRc(result);
      
      
      DemoLogf("Done\n");
    }
    break;
#endif

  case 'D':
    {
      char szPrgfile[255];
      char szEntryPoint[255];

      DemoLogf("Run CodecEngineServer\n");

      DemoLogf("Enter DMP Application file to be loaded (located on DMP target file system): ");
      OsaConsoleGetStr(szPrgfile); 

      DemoLogf("Enter label for the entry point of the application: ");
      OsaConsoleGetStr(szEntryPoint); 

      result = OtxDmpRunCodecEngineServer(g_hDmp[g_demoEnv.nDmp], szPrgfile, szEntryPoint); DemoCheckRc(result);
      DemoLogf("Done\n");
    }
    break;

  case '#':
    {
      DemoLogf("Restart CodecEngineServer\n");

      result = OtxDmpRestartCodecEngineServer(g_hDmp[g_demoEnv.nDmp]); DemoCheckRc(result);
      DemoLogf("Done\n");
    }
    break;

  case 'N':
    {
      char strBuf[255];
      OTX_BOOL bValidMacAddress = OTX_FALSE;
      OTX_UINT8 macAddrBuf[6];
      OTX_UINT8 ipAddrBuf[4];
      OTX_UINT8 netmaskBuf[4];
      char *temp;
      OTX_INT32 i;

      DemoLogf("Set network parameters of the DMP\n");

ENTER_MAC_ADDRESS:
      DemoLogf("Enter MAC Addr [xx.xx.xx.xx.xx.xx]: (or press Enter to skip) ");
      OsaConsoleGetStr(strBuf); 
      if (strBuf[0]) {
        
        temp = strtok(strBuf, "-. ");
        for (i = 6-1; temp && (i >= 0); i--) {
          macAddrBuf[i] = (OTX_UINT8) strtoul(temp, NULL, 16);
          temp = strtok(NULL, "-. ");
        }
        
        if (i != -1) {
          // We did not pick up all the bytes for the mac address. Try again
          DemoLogf("Invalid entry. Try again... \n");
          goto ENTER_MAC_ADDRESS;
        }

        bValidMacAddress = OTX_TRUE;
      }

ENTER_IP_ADDRESS:
      DemoLogf("Enter IP Addr [d.d.d.d]: ");
      OsaConsoleGetStr(strBuf); 
        
      temp = strtok(strBuf, "-. ");
      for (i = 4-1; temp && (i >= 0); i--) {
        ipAddrBuf[i] = (OTX_UINT8) strtoul(temp, NULL, 0);
        temp = strtok(NULL, "-. ");
      }
      
      if (i != -1) {
        // We did not pick up all the bytes for the ip address. Try again
        DemoLogf("Invalid entry. Try again... \n");
        goto ENTER_IP_ADDRESS;
      }

ENTER_NETMASK:
      DemoLogf("Enter Subnet Mask [d.d.d.d]: ");
      OsaConsoleGetStr(strBuf); 
        
      temp = strtok(strBuf, "-. ");
      for (i = 4-1; temp && (i >= 0); i--) {
        netmaskBuf[i] = (OTX_UINT8) strtoul(temp, NULL, 0);
        temp = strtok(NULL, "-. ");
      }
      
      if (i != -1) {
        // We did not pick up all the bytes for the subnet mask. Try again
        DemoLogf("Invalid entry. Try again... \n");
        goto ENTER_NETMASK;
      }

      result = OtxDmpSetNetworkParams(g_hDmp[g_demoEnv.nDmp], (bValidMacAddress) ? macAddrBuf : NULL, ipAddrBuf, netmaskBuf); DemoCheckRc(result);
#if 0
      // Wait for OTP link to become initilized again...
      {
        OTX_UINT8  sqMacAddress[6]; 
        OTX_UINT8  sqIpAddress[4];  
        OTX_UINT8  sqSubnetMask[4]; 
        
        DemoLogf("Waiting for DMP<->HOST link to be initialized ");
        for (i = 0; i < g_demoEnv.nLinkTimeout; i++) {
          result = OtxDmpGetNetworkParams(g_hDmp[g_demoEnv.nDmp], sqMacAddress, sqIpAddress, sqSubnetMask); 
          if (OTX_SUCCESS(result)) {
            DemoLogf("\n");
            DemoLogf("DMP[%d] MAC  =%.2X:%.2X:%.2X:%.2X:%.2X:%.2X:\n", g_demoEnv.nDmp, sqMacAddress[5], sqMacAddress[4], sqMacAddress[3], sqMacAddress[2], sqMacAddress[1], sqMacAddress[0]);
            DemoLogf("DMP[%d] IP   =%d.%d.%d.%d\n",                    g_demoEnv.nDmp, sqIpAddress[3],  sqIpAddress[2],  sqIpAddress[1],  sqIpAddress[0]);
            DemoLogf("DMP[%d] MASK =%d.%d.%d.%d\n",                    g_demoEnv.nDmp, sqSubnetMask[3], sqSubnetMask[2], sqSubnetMask[1], sqSubnetMask[0]);
            break;
          } else {
            OsaThreadSleep(1000);
            DemoLogf(".");
          }
        }
        if (i >= g_demoEnv.nLinkTimeout) {
          DemoLogf("Timeout waiting for DMP<->HOST link to be initialized\n");
          DemoExitQuestion();
        }
      }
      DemoLogf("Done.\n");
#else
      DemoLogf("The system should be rebooted now.\n");
#endif
    }
    break;



  case 'e':
    {
      OTX_UINT32 nProcessId;
      char strBuf[255];
      OTX_RESULT result;

      DemoLogf("Run Linux command (as a standalone process)\n");
      DemoLogf("Enter Linux command: ");

      OsaConsoleGetStr(strBuf); 
      result = OtxDmpRunStandaloneProcess(g_hDmp[g_demoEnv.nDmp], strBuf, &nProcessId); DemoCheckRc(result);

      DemoLogf("Done. %s is now running with PID=%d\n", strBuf, nProcessId);
    }
    break;

  case 'E':
    {
      static OTX_INT32 nProcessId = 0;
      char szProcessName[255];
      char szBuf[255];
      static OTX_INT32 bKillAll = 0;
      OTX_RESULT result;

      DemoLogf("Kill Linux process\n");
      DemoLogf("Enter process name: ");
      OsaConsoleGetStr(szProcessName); 

      DemoLogf("Enter Process ID [%d]: ", nProcessId);
      OsaConsoleGetStr(szBuf); 
      if (szBuf[0]) {
        nProcessId = strtol(szBuf, NULL, 0);
      }

      DemoLogf("Enter 1 to kill all processes with the supplied process name [%d]:  ", bKillAll);
      OsaConsoleGetStr(szBuf); 
      if (szBuf[0]) {
        bKillAll = strtol(szBuf, NULL, 0);
      }

      result = OtxDmpKillStandaloneProcess(g_hDmp[g_demoEnv.nDmp], szProcessName, nProcessId, (OTX_BOOL)bKillAll); DemoCheckRc(result);

      DemoLogf("Done. %s (PID=%d, bKillAll=%d) is now killed\n", szProcessName, nProcessId, bKillAll);
    }
    break;



  case 'x':
    {
      char szHostFileName[255];
      char szTargetPath[255];
      OTX_RESULT result;
      clock_t startTime, stopTime;

      DemoLogf("Upload file to the DMP\n");
      DemoLogf("Enter host file name: ");
      OsaConsoleGetStr(szHostFileName); 

      DemoLogf("Enter target path: ");
      OsaConsoleGetStr(szTargetPath); 

      startTime = clock();
      result = OtxDmpUploadFile(g_hDmp[g_demoEnv.nDmp], szHostFileName, szTargetPath); DemoCheckRc(result);
      stopTime = clock();

      DemoLogf("Done. File %s now uploaded to the %s path (time=%2.1f seconds)\n", szHostFileName, szTargetPath, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;

  case 'X':
    {
      char szHostFileName[255];
      char szTargetPath[255];
      OTX_RESULT result;
      clock_t startTime, stopTime;

      DemoLogf("Download file from the DMP\n");
      DemoLogf("Enter host file name: ");
      OsaConsoleGetStr(szHostFileName); 

      DemoLogf("Enter target path: ");
      OsaConsoleGetStr(szTargetPath); 

      startTime = clock();
      result = OtxDmpDownloadFile(g_hDmp[g_demoEnv.nDmp], szHostFileName, szTargetPath); DemoCheckRc(result);
      stopTime = clock();

      DemoLogf("Done. File %s now downloaded from the %s path (time=%2.1f seconds)\n", szHostFileName, szTargetPath, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;


  case '^':
    {
      char szBuf[255];
      char szHostFileName[255];
      char *szHostFileName2 = "testfile.tmp";
      char *szTargetPath = "testfile.bin";
      OTX_RESULT result;
      int i;
      static int nLoopTimes = 1;
      clock_t startTime, stopTime;

      DemoLogf("File upload/download test\n");
      DemoLogf("Enter host file name to upload and download: ");
      OsaConsoleGetStr(szHostFileName); 

      DemoLogf("Enter number of time to repeat this test [%d]: ", nLoopTimes);
      OsaConsoleGetStr(szBuf); 
      if (szBuf[0]) {
        nLoopTimes = strtol(szBuf, NULL, 0);
      }

      for (i = 0; i < nLoopTimes; i++) {
        DemoLogf("Loop %d of %d\n", i, nLoopTimes);
        
        DemoLogf("Uploading file to the DMP\n");
        startTime = clock();
        result = OtxDmpUploadFile(g_hDmp[g_demoEnv.nDmp], szHostFileName, szTargetPath); DemoCheckRc(result);
        stopTime = clock();
        
        DemoLogf("Done. File %s now uploaded to the %s path (time=%2.1f seconds)\n", szHostFileName, szTargetPath, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
        
        DemoLogf("Downloading file from the DMP\n");
        
        startTime = clock();
        result = OtxDmpDownloadFile(g_hDmp[g_demoEnv.nDmp], szHostFileName2, szTargetPath); DemoCheckRc(result);
        stopTime = clock();
        
        DemoLogf("Done. File %s now downloaded from the %s path (time=%2.1f seconds)\n", szHostFileName2, szTargetPath, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
        
        DemoLogf("Comparing files\n");
        {
#define DEMO_FILE_TEST_BUF_LEN 8192
          FILE *f1;
          FILE *f2;
          char buf1[DEMO_FILE_TEST_BUF_LEN];
          char buf2[DEMO_FILE_TEST_BUF_LEN];
          int coF1;
          int coF2;
          
          f1 = fopen(szHostFileName, "rb");
          if (!f1) {
            DemoLogf("Error: Could not open file %s\n", szHostFileName);
            DemoExitQuestion();
          }
          f2 = fopen(szHostFileName2, "rb");
          if (!f2) {
            DemoLogf("Error: Could not open file %s\n", szHostFileName2);
            DemoExitQuestion();
          }
          
          while ((!feof(f1) && (!feof(f2)))) {
            
            coF1 = fread(buf1, sizeof(char), DEMO_FILE_TEST_BUF_LEN, f1);
            coF2 = fread(buf2, sizeof(char), DEMO_FILE_TEST_BUF_LEN, f2);
            
            if (coF1 != coF2) {
              DemoLogf("Error: Could not read an equal amount of bytes from both files (%d, %d). Files differ.\n", coF1, coF2);
              DemoExitQuestion();
            }
            
            if (memcmp(buf1, buf2, coF1)) {
              DemoLogf("Error: File content differed.\n");
              DemoExitQuestion();
            }
          }
          
          if (feof(f1) && feof(f2)) {
            DemoLogf("Files match\n");
          } else {
            DemoLogf("Error: File size differed.\n");
            DemoExitQuestion();
          }
          
          fclose(f1);
          fclose(f2);
          
          #if defined(WIN32)
          unlink(szHostFileName2);
          #endif
        }

      }
      DemoLogf("File upload/download test completed.\n");
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_RESULT result;

      DemoLogf("Read from a register in the DMP device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg32(g_hDmp[g_demoEnv.nDmp], nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("DMP Address 0x%.4X = 0x%.8X\n", nAddress, (OTX_UINT32)readVal);
    }
    break;

  case 'W':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_RESULT result;

      DemoLogf("Write to a register in the DMP device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg32(g_hDmp[g_demoEnv.nDmp], nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote 0x%.8X to DMP Address 0x%.4X\n", nValue, nAddress);
    }
    break;

  case 'J':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_RESULT result;

      DemoLogf("Read from a register in the g_hBoard device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg32(g_hBoard, nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("hBoard Address 0x%.4X = 0x%.8X\n", nAddress, (OTX_UINT32)readVal);
    }
    break;

  case 'K':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_RESULT result;

      DemoLogf("Write to a register in the g_hBoard device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg32(g_hBoard, nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote 0x%.8X to hBoard Address 0x%.4X\n", nValue, nAddress);
    }
    break;


#if 1 // Command used for debugging only
  case 'a':
    {
      char strBuf[12];
      static OTX_UINT32 nCmd = 0;

      DemoLogf("Debug command for device g_hBoard 0x%X\n", g_hBoard);

      DemoLogf("Enter command (%d) 0=no command sent\n", nCmd);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nCmd = strtoul(strBuf, NULL, 0);
      }
      if(nCmd) {
        OtxDrvDbgDump(g_hBoard, nCmd);
      }
      DemoLogf("Done\n");
    }
    break;
  case 't':
    {
      DemoLogf("Transmit a UPD packet(destport:11000, sourceport:11000) to DMP0\n");
      OtxDrvDbgDump(g_hEthSw, 5);
      DemoLogf("Done\n");
    }
    break;
  case 'T':
    {
      DemoLogf("Transmit a UPD packet to the GMAC port\n");
      OtxDrvDbgDump(g_hEthSw, 1);
      DemoLogf("Done\n");
    }
    break;
  case '*':
    {
      DemoLogf("Configure EthSw from scratch\n");
      result = OtxDrvReset(g_hEthSw); DemoCheckRc(result);
      DemoLogf("Done\n");
    }
    break;
  case 'L':
    {
      DemoLogf("Turn LED D3 ON\n");
      OtxDrvDbgDump(g_hAsm, 4);
    }
    break;
  case 'l':
    {
      DemoLogf("Turn LED D3 OFF\n");
      OtxDrvDbgDump(g_hAsm, 5);
    }
    break;
  case '-':
    {
      OTX_UINT32 nIrpQueueLimit;
      OTX_UINT32 nIrpQueueSize;
      // Print IRP Queue limit and size
      result = OtxDrvGetAttributeValueUint32(g_hDmp[g_demoEnv.nDmp], OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, &nIrpQueueLimit); DemoCheckRc(result);
      DemoLogf("g_hDmp[%d]: nIrpQueueLimit = %d\n", g_demoEnv.nDmp, nIrpQueueLimit);
      result = OtxDrvGetAttributeValueUint32(g_hDmp[g_demoEnv.nDmp], OTX_ATTR_DEVICE_IRP_QUEUE_SIZE, &nIrpQueueSize); DemoCheckRc(result);
      DemoLogf("g_hDmp[%d]: nIrpQueueSize  = %d\n", g_demoEnv.nDmp, nIrpQueueSize);

      result = OtxDrvGetAttributeValueUint32(g_hEthSw, OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, &nIrpQueueLimit); DemoCheckRc(result);
      DemoLogf("g_hEthSw: nIrpQueueLimit = %d\n", nIrpQueueLimit);
      result = OtxDrvGetAttributeValueUint32(g_hEthSw, OTX_ATTR_DEVICE_IRP_QUEUE_SIZE, &nIrpQueueSize); DemoCheckRc(result);
      DemoLogf("g_hEthSw: nIrpQueueSize  = %d\n", nIrpQueueSize);
    }
    break;
  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_RESULT result;

      DemoLogf("Read from an indirect register in the Ethernet Switch device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hEthSw, nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("EthSw Address 0x%.4X = 0x%.2X\n", nAddress, (OTX_UINT8)readVal);
    }
    break;
  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_RESULT result;

      DemoLogf("Write to an indirect register in the Ethernet Switch device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hEthSw, nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote 0x%.2X to EthSw Address 0x%.4X\n", (OTX_UINT8)nValue, nAddress);
    }
    break;
  case 'i':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_RESULT result;

      DemoLogf("Read from a direct register in the Ethernet Switch device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hEthSw, nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("EthSw Address 0x%.4X = 0x%.2X\n", nAddress, (OTX_UINT8)readVal);
    }
    break;
  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_RESULT result;

      DemoLogf("Write to a direct register in the Ethernet Switch device\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hEthSw, nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote 0x%.2X to EthSw Address 0x%.4X\n", (OTX_UINT8)nValue, nAddress);
    }
    break;
  case 'C':
    {
      OTX_UINT8 highWay;
      OTX_UINT8 channel;
      OTX_UINT8 byteVal;
      char wValStr[12];
      OTX_RESULT result;
      
      DemoLogf("Set Constant Byte on Channel\n");
      
      DemoLogf("Highway: ");
      OsaConsoleGetStr(wValStr);
      highWay = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Channel: ");
      OsaConsoleGetStr(wValStr);
      channel = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Constant Byte: ");
      OsaConsoleGetStr(wValStr);
      byteVal = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      result = OtxBrdTssConstByte(g_hTss, highWay, channel, byteVal); 
      DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;
  case '{':
    {     
      char szBuf[12];
      OTX_UINT32 nCmd;
      
      DemoLogf("TSS Debug Command. Which Cmd? ");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_UINT32) strtol(szBuf, NULL, 0);

      DemoLogf("Issuing Driver Debug Command %d to TSS\n", nCmd);
      OtxDrvDbgDump(g_hTss, nCmd);
    }
    break;
  case '!':
    {
      OTX_INT32 nPort;
      OTX_INT32 c;
      OtxEthStatCountersS sStatCounters;
      
      DemoLogf("Get Statistical Counters for Ethernet Switch port. Press port number.\n");
      c = OsaConsoleGetCh();

      if (c >= '0' && c <= '9') {
        nPort = c - '0';
        result = OtxEthGetStatCounters(g_hEthSw, nPort, &sStatCounters); DemoCheckRc(result);
        
        DemoLogf("Statistical Counters for Ethernet Switch port %d\n", nPort);
        DemoPrintStatCounters(&sStatCounters);
      } else {
        DemoLogf("Error: Invalid port number (%c). Valid port range is 0-9.\n", c); 
      }
    }
    break;
  case '|':
    {
      OTX_UINT32 ledSel2regVal;

      DemoLogf("Blinking all LEDs\n");
      OtxDrvWriteReg32(g_hEthSw, 0x1C, 0x3800); // Read operation shadow value prep
      OtxDrvReadReg32(g_hEthSw, 0x1C, &ledSel2regVal); 
      DemoLogf("ledSel2regVal Value = 0x%.4X \n", ledSel2regVal);

      // Temporarily change the function of LED4 since the default function (Intr/Energy) will override the All On/Off bit
      OtxDrvWriteReg32(g_hEthSw, 0x1C, 0xB833); // E=Off

      // Turn all LEDs on 
      OtxDrvWriteReg32(g_hEthSw, 0x10, 0x0010);
      OsaThreadSleep(1500); 
      // Turn all LEDs off
      OtxDrvWriteReg32(g_hEthSw, 0x10, 0x0008);
      OsaThreadSleep(1500); 
      // Turn all LEDs on 
      OtxDrvWriteReg32(g_hEthSw, 0x10, 0x0010);
      OsaThreadSleep(1500); 
      // Turn all LEDs off
      OtxDrvWriteReg32(g_hEthSw, 0x10, 0x0008);
      OsaThreadSleep(1500); 

      // Restore the functions for LED 4 (and LED3)
      OtxDrvWriteReg32(g_hEthSw, 0x1C, ledSel2regVal | 0x8000); 
      
      // Resume normal operation
      OtxDrvWriteReg32(g_hEthSw, 0x10, 0x0000);
    }
    break;

#endif // Debugging commands


   case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;


  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /ME            : E1 Li mode\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /N<num>        : DMP Number (0-3) Default is DMP%d.\n", DEMO_DEFAULT_DMP_NO);
  DemoLogf(" /H             : Longhaul mode (default is shorthaul mode)\n");
  DemoLogf(" /F[E|S|9]      : T1 Frame Format, where\n");
  DemoLogf("                    E = ESF (default)\n");
  DemoLogf("                    4 = F4\n");
  DemoLogf("                    S = SF [D3/D4] (F12)\n");
  DemoLogf("                    9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]        : T1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]      : E1 Frame Format, where\n");
  DemoLogf("                    D = Doubleframe\n");
  DemoLogf("                    C = CRC4 Multiframe (default)\n");
  DemoLogf("                    G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]        : E1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    H = HDB3 (default)\n");
  DemoLogf(" /T[P|4|T|G|9|D] : Board Type used for the demo, where\n");
  DemoLogf("                    P  = Use Thor-8-PCI-Plus Board (default)\n");
  DemoLogf("                    4  = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    T  = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    9  = Use Gimle-16-PCIe Board\n");
  DemoLogf("                    G  = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf("                    D  = Use Alvis-PCI-Express Board\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nDmp              = DEMO_DEFAULT_DMP_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->nLineLength       = 10;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi              = 8;
#if 0
  envP->nIpAddrLsb        = 100;
  envP->nMacAddrLsb       = 0x02;
#else
  envP->nIpAddrLsb        = 111;
  envP->nMacAddrLsb       = 0x52;
#endif
  envP->nLinkTimeout      = 5; // in seconds


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'H') {   // Longhaul mode
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'N') {   // DMP number (0-3)
          envP->nDmp = (OTX_INT32)strtol(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if (*(argv[i] + 2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi        = 8;
          } else if (*(argv[i] + 2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;          
            envP->coLi        = 4;
          } else if (*(argv[i] + 2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;          
            envP->coLi        = 2;
          } else if (*(argv[i] + 2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_ALVIS_PCI_EXPRESS;          
            envP->coLi        = 0;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;          
            envP->coLi        = 16;
          } else if (*(argv[i] + 2) == '9') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE;          
            envP->coLi        = 16;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}



/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  char commandChar;
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; 

  printf("DMP Demo for Alvis-ASM, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen("AlvisDmp.log", "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", "AlvisDmp.log");
    OsaConsoleExit();
		exit(-1);
  }

  result = OtxDrvConnectLib(); 
  DemoCheckRc(result);

  // Create the event queue to be used by the driver to inform the application
  // about events, such as line status changes, or Ethernet switch events, etc.
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }
  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);


  // Open board
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hBoard); 
  DemoCheckRc(result);

  // Read Serial/Build/CPLD Numbers from the board
  {
    char szBuf[64];

    result = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, szBuf, 64); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", szBuf);
    result = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_BUILD, szBuf, 64); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", szBuf);
    result = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_CPLD_BUILD, szBuf, 64); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", szBuf);
  }

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by board)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hBoard, OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < g_demoEnv.coLi/4; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hBoard, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }
    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hLi[i]); DemoCheckRc(result);
    }
    break;
  }

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    // Set default configuration data
    DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

    // Configure E1/T1 Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {

      // Set termination for the T1/E1 Front End
      result = OtxT1E1SetLineTermination(g_hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        result = OtxT1SetConfiguration(g_hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
      } else {
        // E1 mode
        result = OtxE1SetConfiguration(g_hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
      }
      result = OtxT1E1SetLiMode(g_hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);
    }

    // Get a handle to to the ASM Daughter Board
    result = OtxDrvOpenPhysicalDevice(g_hBoard, OTX_DEVICE_ALVIS_ASM, 0, 0, g_hEventQueue, &g_hAsm); DemoCheckRc(result);

    // Open the Ethernet switch
    result = OtxDrvOpenPhysicalDevice(g_hAsm, OTX_DEVICE_ETHERNET_SWITCH, 0, 0, g_hEventQueue, &g_hEthSw); DemoCheckRc(result);
    DemoLogf("OTX_DEVICE_ETHERNET_SWITCH = 0x%X\n", g_hEthSw);

    // Open 4 DMPs
    for (i = 0; i < 4; i++) {
		  result = OtxDrvOpenPhysicalDevice(g_hAsm, OTX_DEVICE_DMP, i, i, g_hEventQueue, &g_hDmp[i]); DemoCheckRc(result);

      DemoLogf("OTX_DEVICE_DMP[%d] = 0x%X\n", i, g_hDmp[i]);
	  }
    break;
  case OTX_DEVICE_ALVIS_PCI_EXPRESS:
    // Open the Ethernet switch
    result = OtxDrvOpenPhysicalDevice(g_hBoard, OTX_DEVICE_ETHERNET_SWITCH, 0, 0, g_hEventQueue, &g_hEthSw); DemoCheckRc(result);
    DemoLogf("OTX_DEVICE_ETHERNET_SWITCH = 0x%X\n", g_hEthSw);

    // Open 8 DMPs
    for (i = 0; i < 8; i++) {
		  result = OtxDrvOpenPhysicalDevice(g_hBoard, OTX_DEVICE_DMP, i, i, g_hEventQueue, &g_hDmp[i]); DemoCheckRc(result);

      DemoLogf("OTX_DEVICE_DMP[%d] = 0x%X\n", i, g_hDmp[i]);
	  }
    break;
  }

  // Get a handle to the Time-Space Switch (TSS) on the board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hBoard, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hTss); DemoCheckRc(result);

  result = OtxDrvEnable(g_hBoard,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  if (g_demoEnv.nBoardType != OTX_DEVICE_ALVIS_PCI_EXPRESS) {
    result = OtxDrvEnable(g_hAsm,    OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

    switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_4_PCI_PLUS:
    case OTX_DEVICE_THOR_8_PCI_PLUS:
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    case OTX_DEVICE_GIMLE_16_PCIE:
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < g_demoEnv.coLi/4; nQuadE1T1++) {
        result = OtxDrvEnable(g_hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }
      break;
    }

    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvEnable(g_hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

      // Remove IDLE channel stuff
      DemoLogf("Removing IDLE channel for Li[%d]\n", i);
      result = OtxDrvWriteReg8(g_hLi[i], 48, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hLi[i], 49, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hLi[i], 50, 0x00); DemoCheckRc(result);
      result = OtxDrvWriteReg8(g_hLi[i], 51, 0x00); DemoCheckRc(result);

    }  
  }

  result = OtxDrvEnable(g_hEthSw, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);



#if 1
  // Connect Pipes
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    case OTX_DEVICE_GIMLE_16_PCIE:
    {
      OTX_UINT8 sqLiTimeSlotMask[32];
      OTX_UINT8 sqDmpTimeSlotMask[128];
      OTX_INT32 nLi;
      OTX_INT32 nDmp = g_demoEnv.nDmp;
      OTX_INT32 nTsLi;
      OTX_INT32 nTsDmp;
      OTX_HANDLE hPipe;

      for (nLi = 0; nLi < 4; nLi++) {
        for (nTsLi = 1; nTsLi < 32; nTsLi++) {
          
          memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqLiTimeSlotMask[nTsLi] = 0xFF;
          
          nTsDmp = nTsLi + nLi * 32;
          memset(sqDmpTimeSlotMask, 0, 128 * sizeof(OTX_UINT8));
          sqDmpTimeSlotMask[nTsDmp] = 0xFF;
          
          // Create a pipe for each timeslot  (Li -> DMP)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 128); DemoCheckRc(result);

        }
      }
      DemoLogf("Pipes connected from DMP[%d] to Li0-Li3\n", g_demoEnv.nDmp);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      OTX_UINT8 sqLiTimeSlotMask[32];
      OTX_UINT8 sqDmpTimeSlotMask[128];
      OTX_INT32 nLi;
      OTX_INT32 nDmp = g_demoEnv.nDmp;
      OTX_INT32 nTsLi;
      OTX_INT32 nTsDmp;
      OTX_HANDLE hPipe;
      OTX_UINT32 nDmpHighwaySpeed;
      
        // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hDmp[nDmp], OTX_ATTR_DMP_HIGHWAY_SPEED, &nDmpHighwaySpeed); 
      DemoLogf("DMP#%d highway speed is %d kbps\n", nDmp, nDmpHighwaySpeed);
      
      for (nLi = 0; nLi < 4; nLi++) {
        for (nTsLi = 1; nTsLi < 32; nTsLi++) {
          
          memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqLiTimeSlotMask[nTsLi] = 0xFF;
          
          nTsDmp = nTsLi + nLi * 32;
          memset(sqDmpTimeSlotMask, 0, 128 * sizeof(OTX_UINT8));
          sqDmpTimeSlotMask[nTsDmp] = 0xFF;
          
          // Create a pipe for each timeslot  (Li -> DMP)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 128); DemoCheckRc(result);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);

          // Create a pipe for each timeslot  (DMP -> Li)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 128); DemoCheckRc(result);

        }
      }
      DemoLogf("Pipes connected from DMP[%d] to Li0-Li3\n", g_demoEnv.nDmp);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      OTX_UINT8 sqLiTimeSlotMask[32];
      OTX_UINT8 sqDmpTimeSlotMask[64];
      OTX_INT32 nLi;
      OTX_INT32 nDmp = g_demoEnv.nDmp;
      OTX_INT32 nTsLi;
      OTX_INT32 nTsDmp;
      OTX_HANDLE hPipe;
      
      for (nLi = 0; nLi < 1; nLi++) {
        for (nTsLi = 1; nTsLi < 32; nTsLi++) {
          
          memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqLiTimeSlotMask[nTsLi] = 0xFF;
          
          nTsDmp = nTsLi + nLi * 32;
          memset(sqDmpTimeSlotMask, 0, 64 * sizeof(OTX_UINT8));
          sqDmpTimeSlotMask[nTsDmp] = 0xFF;
          
          // Create a pipe for each timeslot  (Li -> DMP)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 64); DemoCheckRc(result);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);

          // Create a pipe for each timeslot  (DMP -> Li)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 64); DemoCheckRc(result);

        }
      }
      DemoLogf("Pipes connected from DMP[%d] to Li0\n", g_demoEnv.nDmp);
    }
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    {
      OTX_UINT8 sqLiTimeSlotMask[32];
      OTX_UINT8 sqDmpTimeSlotMask[32];
      OTX_INT32 nLi;
      OTX_INT32 nDmp = g_demoEnv.nDmp;
      OTX_INT32 nTsLi;
      OTX_INT32 nTsDmp;
      OTX_HANDLE hPipe;
      
      for (nLi = 0; nLi < 1; nLi++) {
        for (nTsLi = 1; nTsLi < 32; nTsLi++) {
          
          memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqLiTimeSlotMask[nTsLi] = 0xFF;
          
          nTsDmp = nTsLi + nLi * 32;
          memset(sqDmpTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
          sqDmpTimeSlotMask[nTsDmp] = 0xFF;
          
          // Create a pipe for each timeslot  (Li -> DMP)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 32); DemoCheckRc(result);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);

          // Create a pipe for each timeslot  (DMP -> Li)
          result = OtxDrvCreatePipe(64, &hPipe);
          //DemoLogf("Connecting: nLi=%d, nTsLi=%d\n", nLi, nTsLi);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hLi[nLi], -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
          //DemoLogf("Connecting: nDmp=%d, nTsDmp=%d\n", nDmp, nTsDmp);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hDmp[nDmp], -1, sqDmpTimeSlotMask, 32); DemoCheckRc(result);

        }
      }
      DemoLogf("Pipes connected from DMP[%d] to Li0\n", g_demoEnv.nDmp);
    }
    break;
  }
#endif

  OsaThreadStart(DemoIdleFunction);

  // Print the network parameters for DMP[g_demoEnv.nDmp]
  {
    OTX_UINT8  sqMacAddress[6];  // MAC Address of DMP. Provide a NULL parameter to leave the MAC address unchanged. 6 bytes. Example 00:50:C2:81:70:0B. sqEthDevice[5] is the 0x00 byte (MSB) and sqEthDevice[0] is the 0x0B byte (LSB)
    OTX_UINT8  sqIpAddress[4];   // IP Address of DMP Ethernet port. 4 bytes. Example 10.0.1.103. sqDmpIpAddr[3] is the 10 byte (MSB) and sqDmpIpAddr[0] is the 103 byte (LSB)
    OTX_UINT8  sqSubnetMask[4];  // Subnet mask of DMP Ethernet port. 4 bytes. Example 10.0.1.200. sqDmpIpAddr[3] is the 10 byte (MSB) and sqDmpIpAddr[0] is the 200 byte (LSB)

    DemoLogf("Waiting up to %d seconds for DMP<->HOST link to be initialized...\n", g_demoEnv.nLinkTimeout);
    for (i = 0; i < g_demoEnv.nLinkTimeout; i++) {
      result = OtxDmpGetNetworkParams(g_hDmp[g_demoEnv.nDmp], sqMacAddress, sqIpAddress, sqSubnetMask); 
      if (OTX_SUCCESS(result)) {
        DemoLogf("\n");
        DemoLogf("DMP[%d] MAC  =%.2X:%.2X:%.2X:%.2X:%.2X:%.2X:\n", g_demoEnv.nDmp, sqMacAddress[5], sqMacAddress[4], sqMacAddress[3], sqMacAddress[2], sqMacAddress[1], sqMacAddress[0]);
        DemoLogf("DMP[%d] IP   =%d.%d.%d.%d\n",                    g_demoEnv.nDmp, sqIpAddress[3],  sqIpAddress[2],  sqIpAddress[1],  sqIpAddress[0]);
        DemoLogf("DMP[%d] MASK =%d.%d.%d.%d\n",                    g_demoEnv.nDmp, sqSubnetMask[3], sqSubnetMask[2], sqSubnetMask[1], sqSubnetMask[0]);
        break;
      } else {
        OsaThreadSleep(1000);
        DemoLogf(".\n");
      }
    }
    if (i >= g_demoEnv.nLinkTimeout) {
      DemoLogf("\nTimeout waiting for DMP<->HOST link to be initialized\n");
      //DemoExitQuestion();
    } else {
      // Read Serial Number from the DMM board (via the DMP<->HOST link)
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      char *pszSerialNumber;
      
      result = OtxDrvGetAttributeType(g_hDmp[g_demoEnv.nDmp], OTX_ATTR_DMP_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      if (cbAttribValue) {
        pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
        result = OtxDrvGetAttributeValueSz(g_hDmp[g_demoEnv.nDmp], OTX_ATTR_DMP_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
        DemoLogf("DMM Serial Number: %s\n", pszSerialNumber);
        free(pszSerialNumber);
      }
    }
  }



  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }


  printf("Complete\n");
  OtxDrvReleaseLib();

  return 0;
}

